﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using Ventuz.Remoting2;
using Ventuz.Remoting2.Helpers;

namespace RemoteClient
{
    /// <summary>
    /// The Render To Disk GUI.
    /// This application gives an example how to use the new IRenderToDisk interface (via the RemoteRenderToDisk helper class) to
    /// create an image sequence of a Ventuz scene.
    /// </summary>
    public partial class Render2DiskForm : Form
    {
        #region enum CaptureStatus
        
        [Flags]
        enum CaptureStatus
        {
            None = 0,
            PreparingScene = 1,
            Recording = 2,
            RenderToDisk = 8,
            CommandRecordingPossible = 0x0F0,
            RenderToDiskPossible = 0x100,

            ProgressMask = 0x00F,
            StatusMask = 0xFFF0,
        } 

        #endregion

        #region members

        PlayoutConfig config = null;                            // currently just used internally and thus not really necessary
        CaptureStatus captureStatus = CaptureStatus.None;       // represents the current capture/recording/playout states
        RemoteRenderToDisk remoteR2D = null;                    // handle to the render-to-disk service
        double prog = 0;                                        // playout progress value for the progress bar

        #endregion

        #region properties
        
        /// <summary>
        /// Timecode (frameindex) set in the GUI. 
        /// </summary>
        public ulong Timecode
        {
            get { return (ulong)this.numUD_TC.Value; }
        }

        /// <summary>
        /// Returns true if currently recording remote commands.
        /// </summary>
        public bool RecordingCommands
        {
            get { return (this.captureStatus & CaptureStatus.Recording) != 0; }
        }

        #endregion


        #region Constructor
        
        public Render2DiskForm(PlayoutConfig config, string ip)
        {
            InitializeComponent();

            // init to saved config
            this.config = config;
            if ( this.config == null )
                this.config = new PlayoutConfig();

            if ( string.IsNullOrEmpty(ip) )
                this.lbl_IP.Text = ip;

            ConfigToUI();

            // disable GUI elements until we connect to Ventuz
            this.btClearCmds.Enabled = false;
            this.tabControl.Enabled = false;
        } 

        #endregion

        #region Initializations
        
        /// <summary>
        /// Configures the GUI according to the parameters saved in the PlayoutConfig class
        /// </summary>
        private void ConfigToUI()
        {
            if ( this.config != null )
            {
                // options 
                this.cbSceneActivate.Checked = this.config.DoNotFireSceneEvents;

                // playout
                this.btRenderToDisk.Enabled = false;

                // playback button
                if ( this.config.CapturedInteraction && (this.captureStatus & CaptureStatus.RenderToDiskPossible) != 0 )
                    this.btClearCmds.Enabled = true;
                else
                    this.btClearCmds.Enabled = false;
                
                // image file format
                this.numUpDownFPS.Value = (decimal)this.config.FPS;
                this.cbAlphaChannel.Checked = this.config.Alpha;
                this.comBoxImageFormat.SelectedIndex = (int)this.config.ImageFormat;

                // output files
                this.tbBaseFilename.Text = this.config.BaseFilename;
                this.tbOutputPath.Text = this.config.OutputFolder;
            }
        }

        /// <summary>
        /// Sets the PlayoutConfig class instance according to the state of the GUI
        /// </summary>
        private void UIToConfig()
        {
            if ( this.config == null )
                this.config = new PlayoutConfig();

            // options
            this.config.DoNotFireSceneEvents = this.cbSceneActivate.Checked;

            // image format
            this.config.FPS = (double)this.numUpDownFPS.Value;
            this.config.Alpha = this.cbAlphaChannel.Checked;
            this.config.ImageFormat = (PlayoutImageFormat)this.comBoxImageFormat.SelectedIndex;

            // output path
            this.config.BaseFilename = this.tbBaseFilename.Text;
            this.config.OutputFolder = this.tbOutputPath.Text;
        }

        /// <summary>
        /// Resets the GUI
        /// </summary>
        private void ResetUI()
        {
            // reset capture status - progress status is set to none
            this.captureStatus &= CaptureStatus.StatusMask;

            // Configure GUI elements:
            // labels
            this.lblCaptureHelpText.Visible = false;
            this.lblCaptureHelpText.Text = "Prepare scene and press Start Recording";

            // buttons
            this.btCaptureSteps.Text = "Start Capture";
            this.btCaptureSteps.ForeColor = SystemColors.ControlText;
            this.btClearCmds.Text = "Clear Recording";

            // progress bar
            this.progBarPlayout.Visible = false;

            // enable buttons depending on capture status
            if ( (this.captureStatus & CaptureStatus.RenderToDiskPossible) != 0 )
            {
                // input records still available
                this.btRenderToDisk.Enabled = true;
                this.btClearCmds.Enabled = true;
            }
            else
            {
                // no input records available
                this.btRenderToDisk.Enabled = false;
                this.btClearCmds.Enabled = false;
            }

            if ( (this.captureStatus & CaptureStatus.CommandRecordingPossible) != 0 )
                this.btCaptureSteps.Enabled = true;
            else
                this.btCaptureSteps.Enabled = false;

            // tabs
            this.tabControl.Enabled = true;
        } 

        #endregion


        #region G U I   E V E N T S

        #region Cancel
        
        /// <summary>
        /// Processes the Cancel button dependent on the active action
        /// </summary>
        private void btCancel_Click(object sender, EventArgs e)
        {
            if ( (this.captureStatus & CaptureStatus.ProgressMask) == CaptureStatus.Recording || (this.captureStatus & CaptureStatus.ProgressMask) == CaptureStatus.PreparingScene )
            {
                // dismiss recorded data
                if ( this.remoteR2D != null )
                    this.remoteR2D.ClearRecordedCommands();

                ResetUI();
            }
            else if ( (this.captureStatus & CaptureStatus.ProgressMask) == CaptureStatus.RenderToDisk )
            {
                // stop playout (render to disk)!
                if ( this.remoteR2D != null )
                    this.remoteR2D.StopPlayoutToDisk();

                // to stop progress-bar-update-loop
                this.prog = 100;

                ResetUI();
            }
            else
            {
                // close all
                if ( this.remoteR2D != null )
                {
                    this.remoteR2D.StopCommandRecording(this.Timecode);
                    this.remoteR2D.StopRenderToDiskMode();
                    this.remoteR2D.ClearRecordedCommands();
                }

                Close();
            }
        }

        #endregion


        #region Capture steps (Prepare/Record/Stop)
        
        /// <summary>
        /// Handles the different steps of a render-to-disk process
        /// </summary>
        private void btCaptureSteps_Click(object sender, EventArgs e)
        {
            UIToConfig();

            switch ( this.captureStatus & CaptureStatus.ProgressMask )
            {
                case CaptureStatus.None:
                case CaptureStatus.CommandRecordingPossible:
                    // configure GUI to 'preparing scene' status
                    this.lblCaptureHelpText.Visible = true;
                    this.btCaptureSteps.Text = "Start Recording";
                    this.btCaptureSteps.ForeColor = Color.Green;
                    this.numUD_TC.Enabled = true;

                    // disable other GUI elements
                    this.btRenderToDisk.Enabled = false;
                    this.btClearCmds.Enabled = false;

                    // START RENDER-TO-DISK MODE
                    // ----
                    this.captureStatus |= CaptureStatus.PreparingScene;
                    try
                    {
                        this.remoteR2D.StartRenderToDiskMode();
                        Console.WriteLine("RemoteR2D - Render2Disk mode started!");
                    }
                    catch ( Exception ex )
                    {
                        string msg = string.Format("Could not start Render-to-disk mode in Ventuz:\n{0}", ex.Message);
                        MessageBox.Show(msg, "Render To Disk");

                        // revert 
                        this.captureStatus &= ~CaptureStatus.PreparingScene;

                        this.lblCaptureHelpText.Visible = false;
                        this.lblCaptureHelpText.Text = "Prepare scene and press Start Recording.";
                        this.btCaptureSteps.Text = "Start Capture";
                        this.btCaptureSteps.ForeColor = SystemColors.ControlText;

                        if ( (this.captureStatus & CaptureStatus.RenderToDiskPossible) != 0 )
                            this.btRenderToDisk.Enabled = true;
                        this.btClearCmds.Enabled = true;
                    }
                    // ----

                    break;

                case CaptureStatus.PreparingScene:
                    // configure GUI to 'recording commands' status
                    this.lblCaptureHelpText.Text = "Teach-in commands. Press Stop Recording if done.";
                    this.btCaptureSteps.Text = "Stop Recording";
                    this.btCaptureSteps.ForeColor = Color.Red;
                    this.btClearCmds.Enabled = true;

                    // START COMMAND RECORDING
                    // ----
                    this.captureStatus = (this.captureStatus & CaptureStatus.StatusMask) | CaptureStatus.Recording;
                    this.remoteR2D.RendererDuringCommandRecording(this.cbRecPreview.Checked);
                    this.remoteR2D.StartCommandRecording(this.Timecode);
                    Console.WriteLine("RemoteR2D - Command Recording started!");
                    // ----

                    break;

                case CaptureStatus.Recording:
                    // configure GUI to 'none' or 'record available' status
                    this.lblCaptureHelpText.Visible = false;
                    this.lblCaptureHelpText.Text = "Prepare scene and press Start Recording.";
                    this.btCaptureSteps.Text = "Start Capture";
                    this.btCaptureSteps.ForeColor = SystemColors.ControlText;

                    // update buttons because we at least now have an commands recording
                    this.btRenderToDisk.Enabled = true;
                    this.numUD_TC.Enabled = false;

                    // UPDATE STATUS AND STOP COMMAND RECORDING
                    // ----
                    this.captureStatus = (this.captureStatus & CaptureStatus.StatusMask) | CaptureStatus.RenderToDiskPossible;
                    this.remoteR2D.StopCommandRecording(this.Timecode);
                    //this.remoteR2D.StopRenderToDiskMode();
                    Console.WriteLine("RemoteR2D - Command Recording stopped!");
                    // ----

                    break;

                default:
                    break;
            }
        }

        #endregion
        
        #region Render to disk

        /// <summary>
        /// Collect the required parameters and start the image sequence playout in Ventuz for the previously recorded remote commands.
        /// </summary>
        private void btRenderToDisk_Click(object sender, EventArgs e)
        {
            if ( (this.captureStatus & CaptureStatus.RenderToDiskPossible) == 0 )
            {
                MessageBox.Show("Render To Disk not possible because there is no Recording available!", "Error");
                return;
            }

            // get current playout configuration
            UIToConfig();

            // update status
            this.captureStatus |= CaptureStatus.RenderToDisk;

            // disable tab switching during playout
            this.tabControl.Enabled = false;

            // disaable other buttons
            this.btCaptureSteps.Enabled = false;
            this.btClearCmds.Enabled = false;

            // show progress bar
            this.progBarPlayout.Enabled = true;
            this.progBarPlayout.Visible = true;
            this.progBarPlayout.Value = 0;

            try
            {
                int fps = (int)this.numUpDownFPS.Value;
                PlayoutImageFormat iFormat = (PlayoutImageFormat)this.comBoxImageFormat.SelectedIndex;
                string output = this.tbOutputPath.Text;
                string bName = this.tbBaseFilename.Text;
                bool keepAlpha = this.cbAlphaChannel.Checked;
                bool sceneEvents = this.cbSceneActivate.Checked;

                try
                {
                    // this is a non-blocking call
                    if ( this.cbProjectRes.Checked )
                    {
                        // use project resolution for capturing
                        this.remoteR2D.StartPlayoutToDisk(fps, false, iFormat, output, bName, keepAlpha, sceneEvents);
                    }
                    else
                    {
                        // custom resolution
                        this.remoteR2D.StartPlayoutToDisk(fps, false, iFormat, PlayoutInterlacing.None, (uint)this.numUpDown_Width.Value, (uint)this.numUpDown_Height.Value,output, bName, keepAlpha, sceneEvents);
                    }

                    this.prog = 0;

                    // update progress bar during playout
                    while ( this.prog < 100.0 )
                    {
                        this.prog = this.remoteR2D.GetPlayoutProgress();

                        this.progBarPlayout.Value = (int)(prog);
                        Application.DoEvents();
                        System.Threading.Thread.Sleep(10);
                    }
                }
                catch ( Exception ex )
                {
                    string msg = string.Format("Playout failed: {0}", ex.Message);
                    MessageBox.Show(msg, "Render To Disk");
                }
            }
            finally
            {
                // REVERT above GUI enable/disable steps
                this.tabControl.Enabled = true;
                this.btCaptureSteps.Enabled = true;
                this.btClearCmds.Enabled = true;
                this.progBarPlayout.Visible = false;         
            }

            this.captureStatus &= ~CaptureStatus.RenderToDisk;
        } 

        #endregion

        #region Clear Recordings

        /// <summary>
        /// Clear recorded remote commands.
        /// </summary>
        private void btClearCmds_Click(object sender, EventArgs e)
        {
            if ( this.remoteR2D != null )
                this.remoteR2D.ClearRecordedCommands();

            this.btClearCmds.Enabled = false;
            this.btRenderToDisk.Enabled = false;
            this.captureStatus &= ~CaptureStatus.RenderToDiskPossible;
        }

        private void PlaybackFinished(object sender, EventArgs e)
        {
            this.btRenderToDisk.Enabled = true;
        }

        #endregion

        #region Preview
        
        /// <summary>
        /// Enable immediate remote command processing on Ventuz site to have a preview/feedback in the Render output
        /// </summary>
        private void cbRecPreview_CheckedChanged(object sender, EventArgs e)
        {
            if ( this.remoteR2D != null )
            {
                this.remoteR2D.RendererDuringCommandRecording(this.cbRecPreview.Checked);
            }
        } 

        #endregion


        #region CONNECT
        
        /// <summary>
        /// Connects to the render-to-disk service of Ventuz
        /// </summary>
        private void btConnect_Click(object sender, EventArgs e)
        {
            try
            {
                this.remoteR2D = RemoteRenderToDisk.Connect(this.tb_Host.Text);
                this.tabControl.Enabled = true;
                this.captureStatus = CaptureStatus.CommandRecordingPossible;

                Console.WriteLine("RemoteR2D - Connected to Ventuz R2D-Service!");
            }
            catch ( Exception ex )
            {
                string msg = string.Format("Connect to Ventuz Render-to-disk Service failed: {0}", ex.Message);
                MessageBox.Show(msg, "Render To Disk");
                this.captureStatus = CaptureStatus.None;
            }
        } 

        #endregion

        private void cbProjectRes_CheckedChanged(object sender, EventArgs e)
        {
            if ( this.cbProjectRes.Checked )
            {
                this.numUpDown_Width.Enabled = false;
                this.numUpDown_Height.Enabled = false;
            }
            else
            {
                this.numUpDown_Width.Enabled = true;
                this.numUpDown_Height.Enabled = true;
            }
        }


        #endregion
    }

    #region class PlayoutConfig
    
    public class PlayoutConfig
    {
        #region Fields

        // Options
        public bool DoNotFireSceneEvents;

        // Source Interactive
        public double CapturedTime;
        public ulong CaptureStartedFrame;
        public ulong CaptureStoppedFrame;
        public bool CapturedInteraction;

        // Image Format
        public double FPS;
        public bool Alpha;
        public PlayoutImageFormat ImageFormat;

        // Output Path
        public string BaseFilename;
        public string OutputFolder;

        #endregion

        #region Constructor / Init

        public PlayoutConfig()
        {
            Init();
        }

        private void Init()
        {
            // Source 
            DoNotFireSceneEvents = true;

            // Source Interactive
            CapturedTime = 0.0;
            CaptureStartedFrame = 0;
            CaptureStoppedFrame = 0;
            CapturedInteraction = true;

            // Image Format
            FPS = 30;
            Alpha = true;
            ImageFormat = PlayoutImageFormat.TIF;

            // Output Path
            BaseFilename = "Ventuz";
            OutputFolder = "<none>";
        }

        #endregion
    } 

    #endregion
}
